/*
program:  for vinet (simulating a LAN for ECO32)
purpose:  main program for the VirtualNetwork process, doing all the actual simulation.
task:     Developed as a project in course "Mikroprozessortechnik II" (simulation of 
          processors) at the University of Applied Sciences Giessen-Friedberg in 
          winter term 2003. The course was led by Prof. Dr. Hellwig Geisse.
authors:  Jochen Mller
          Mathias Schfer
          Mattias Ansorg <matthias@ansorgs.de>
license:  public domain
date:     2003-11-03 to 
usage:    
*/


#include <stdio.h>
#include <rpc/rpc.h>        /* always needed */
#include "ViNet.h"
#include "ViNet_Protocol.h" /* need this too; will be generated by rpcgen */
#include "IfProxy_Protocol.h" /* need this too; will be generated by rpcgen */

/* 
 * dispatch routine is defined in server stub ViNet_Protocol_svc.c generated by rpcgen
 */
extern void nicendprog_1(struct svc_req *, register SVCXPRT *);

/* how to contact peer IfProxy rpc server: client handle for first IfProxy rpc server
   must be global as the interface of routines using it is defined by rpcgen */
CLIENT *cl[10];
int cl_new = 0; /* first free element in cl array */

/* 
 * implementation of routine provided via rpc
 * returns pointer on int with value 0 on success, value nonzero otherwise
 */
int * nicregister_1_svc(char *server, int ifendprog, int ifendvers, struct svc_req *rqstp) {
  static int result; /* must be static! */
  char myhostname[HOSTNAMELEN]; /* name of localhost */

  /* get name of localhost */
  result = gethostname(myhostname, HOSTNAMELEN);
  if (result != 0) {
    fprintf(stderr, "Error: could not get my own hostname.\n");
    fprintf(stderr, "Effect: local loopback device (localhost) will be used in _no_ case.\n");
  }
  
  /* Create client "handle" used for calling the IfProxy peer rpc server that makes itself 
   * known by calling this routine */
  if (strcmp(myhostname,server) == 0) {
    /* using localhost where possible prevents errors if the real network is unreachable */
    cl[cl_new] = clnt_create("localhost", ifendprog, ifendvers, "udp");
  }
  else
    cl[cl_new] = clnt_create(server, ifendprog, ifendvers, "udp");

  if (cl[cl_new] == NULL) {
    /* Couldn't establish connection with server. Print error message, return error code. */
    clnt_pcreateerror(server);
    result = 1;
    return &result;
  }
  else {
    ++cl_new;
    result = 0;
    return &result;
  }
}

/* 
 * implementation of routine provided via rpc
 * returns pointer on int with value 0 on success, value nonzero otherwise
 */
int * broadcastmessage_1_svc(char *msg, struct svc_req *rqstp) {
  static int result; /* must be static! */
  int *result_ptr;
  int i;
  FILE *f;
  
  f = fopen("/dev/stdout", "w");
  if (f == NULL) { 
    result = 1;
    return (&result);
  }
  fprintf(f, "rpc call received: broadcast message \"%s\"\n", msg);
  fclose(f);
  
  for (i=0; i < cl_new; ++i) {
    /* call all IfProxy servers including that of the calling client to print that 
     * same message ("broadcasting"). */
    result_ptr = printmessage_1(msg, cl[i]);
    if (result_ptr == NULL) {
      /* An error occurred while calling the server. Print error message and die. */
      clnt_perror(cl[i], "Error");
      /* no influence on result here as result shapes only the success of local ops */
    }
    else {
      result = *result_ptr;
    }
    /* Okay, we successfully called the remote procedure. */
    if (result != 0) {
      /* Server was unable to print our message. Print error message. */
      fprintf(stderr, "Error: one IfProxy server couldn't print the message \"%s\"\n", msg);
      /* no influence on result here as result shapes only the success of local ops */
    }
    else {
      /* The message got printed on the server's console */ 
      printf("rpc called: print message \"%s\" on IfProxy server\n", msg);
    }
  }
 
  result = 0;
  return &result;
}

/*
 * initialize rpc server provided by ViNet
 * returns 0 on success, 1 otherwise
 */
int svc_init () {
  register SVCXPRT *transp;

  pmap_unset(NICENDPROG, NICENDVERS);

  transp = svcudp_create(RPC_ANYSOCK);
  if (transp == NULL) {
    fprintf (stderr, "cannot create udp service\n");
    return 1;
  }
  if (!svc_register(transp, NICENDPROG, NICENDVERS, nicendprog_1, IPPROTO_UDP)) {
    fprintf (stderr, "unable to register (NICENDPROG, NICENDVERS, udp)\n");
    return 1;
  }

  return 0;
}

/*
 * ViNet main function
 */
int main (int argc, char **argv) {
  /* correct usage of this program? */
  if (argc > 1) {
    fprintf(stderr, "usage: %s", argv[0]);
    exit(1);
  }
  
  /* initialize and run own rpc server */
  if (svc_init() != 0) {
    exit(1);
  }
  svc_run ();
  fprintf (stderr, "svc_run returned\n");
  exit (1); /* NOTREACHED */
}
